package shard

import (
	"context"

	"git.frostfs.info/TrueCloudLab/frostfs-api-go/v2/pkg/tracing"
	"git.frostfs.info/TrueCloudLab/frostfs-node/internal/logs"
	meta "git.frostfs.info/TrueCloudLab/frostfs-node/pkg/local_object_storage/metabase"
	oid "git.frostfs.info/TrueCloudLab/frostfs-sdk-go/object/id"
	"go.opentelemetry.io/otel/attribute"
	"go.opentelemetry.io/otel/trace"
	"go.uber.org/zap"
)

// ToMoveItPrm encapsulates parameters for ToMoveIt operation.
type ToMoveItPrm struct {
	addr oid.Address
}

// ToMoveItRes encapsulates results of ToMoveIt operation.
type ToMoveItRes struct{}

// SetAddress sets object address that should be marked to move into another
// shard.
func (p *ToMoveItPrm) SetAddress(addr oid.Address) {
	p.addr = addr
}

// ToMoveIt calls metabase.ToMoveIt method to mark object as relocatable to
// another shard.
func (s *Shard) ToMoveIt(ctx context.Context, prm ToMoveItPrm) (ToMoveItRes, error) {
	ctx, span := tracing.StartSpanFromContext(ctx, "Shard.ToMoveIt",
		trace.WithAttributes(
			attribute.String("shard_id", s.ID().String()),
			attribute.String("address", prm.addr.EncodeToString()),
		))
	defer span.End()

	s.m.RLock()
	defer s.m.RUnlock()

	m := s.info.Mode
	if m.ReadOnly() {
		return ToMoveItRes{}, ErrReadOnlyMode
	} else if m.NoMetabase() {
		return ToMoveItRes{}, ErrDegradedMode
	}

	var toMovePrm meta.ToMoveItPrm
	toMovePrm.SetAddress(prm.addr)

	_, err := s.metaBase.ToMoveIt(ctx, toMovePrm)
	if err != nil {
		s.log.Debug(logs.ShardCouldNotMarkObjectForShardRelocationInMetabase,
			zap.String("error", err.Error()),
		)
	}

	return ToMoveItRes{}, nil
}
